#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_cache.h>
#include <cyg/infra/diag.h>
#include <stdio.h>

#include <time.h>
#include <stdlib.h>

#include "dm.h"
#include <cyg/hal/iic.h>
#include "ident.h"
#include "watchdog.h"

// product ID info
tIdent ident;

// state for background WatchDog task
unsigned BgStack[1024];
cyg_handle_t  BgHandle;
cyg_thread    BgThread;

// state for diagnostic task
unsigned DiagStack[1024];
cyg_handle_t  DiagHandle;
cyg_thread    DiagThread;

// global flag set by TCP/IP stack init routine
int TcpStackOk;
//??
taskList_t taskList[10]; //  Task monitoring struct

// declare tasks
static cyg_thread_entry_t BgTask;
static cyg_thread_entry_t DiagTask;

/*  Malloc information. */
#include <pkgconf/system.h>
#include CYGHWR_MEMORY_LAYOUT_H
#define HEAP_START     ((unsigned long)(&CYGMEM_SECTION_heap94))
#define HEAP_SIZE      ((unsigned long)(CYGMEM_SECTION_heap94_SIZE))
#define HEAP_END       (HEAP_START + HEAP_SIZE)

/*  Fancy macro for the portmon feature */
CYG_HAL_TABLE_BEGIN( __DeviceMaster_DIAG_TAB__, DeviceMaster_diag );
CYG_HAL_TABLE_END( __DeviceMaster_DIAG_TAB_END__, DeviceMaster_diag );
extern struct diag_tab_entry __DeviceMaster_DIAG_TAB__[], __DeviceMaster_DIAG_TAB_END__;

//Serial Hub 8/16 flag
int HAS_ONLY_232;

/*  
 *  Set the firmware identifying string here.  This shows up on the web page, on the 
 *  debug port, and in Comtrol's PortVision application.
 */
static char firmwareName[] = "DMDK Customer SocketServer";
static int versionMajor = 2;
static int versionMinor = 1;

char versionString[MAX_VERSIONSTRING_SIZE];  /*  The global firmware identification string variable, max size is 64 bytes. */

/*  Prototypes */
static void TcpIpStackInit(void);

// Here is where user execution starts
void cyg_user_start(void)
{
  int i;

  LedEnable();
  LedOff();

  sprintf(versionString,"%s %d.%02d", firmwareName, versionMajor, versionMinor);
  
  // tell the world we're alive
  diag_printf("%s\n",versionString);
  diag_printf("Copyright Comtrol Corp. 2001-2004\n");
  diag_printf("Build date: %s\n\n",linkTimeStamp);
  
  // enable the unified I/D cache in processor
  HAL_DCACHE_ENABLE();
  
  // read id stuff from flash
  getMacAddress(ident.macAddr);
  getIpInfo(&ident.ipAddr, &ident.netMask, &ident.gateway);
  getModelNumber(&ident.modelId);
  getBoardRev(&ident.boardRev);
  getPassword(ident.password);
  getAuthMethod(&ident.authMethod);
  getTelnetEnable(&ident.telnetEnable);
  getTelnetTimeout(&ident.telnetTimeout);
  getRedbootDelay(&ident.timeoutValue);
  
  ident.archId = 3;
  ident.programId = 255;  // somebody else has to set this

  ident.versionMajor = versionMajor;
  ident.versionMinor = versionMinor;
  
   switch (ident.modelId)
   {
   case 5002110:		//RTS 8 DB9
   case 5002112:		//RTS 8 RJ45
   case 5002115:		//SH 8
   case 5002330:		//PRO 8
     ident.numPorts = 8;
       break;
   case 5002111:		//RTS 4 DB9
   case 5002113:		//RTS 4 RJ45
   case 5002117:                //DM500 4 Port DB9
       ident.numPorts = 4;
       break;
   case 5002120:		//RTS 16 RJ45
   case 5002240:                // SPS 16
   case 5002315:		//SH 16 DB9
   case 5002325:		//PRO 16 RJ45
   case 5002116:		// IAD UP 4 Port
       ident.numPorts = 16;
       break;
   case 5002241:                //  SPS 32
       ident.numPorts = 32;
       break;
   case 5002280:                //  DM 1 Port
   case 5002281:                //  DM 1 Port Embedded
   case 5002283:                //  Air 1 Port
   case 5002360:		//  DM 1 Port 5 - 30 VDC
   case 5002361:		//  DM 1 Port 5 - 30 VDC Embedded
   case 5002405:		//DM2 1 port
   case 5002362:                //DM500 1 Port Embedded
   case 5002363:                //DM500 1 Port DB9
       ident.numPorts = 1;
       break;
   case 5002525:                //  DM RTS 2P 1E (Configuration 1)
   case 5002535:                //  DM RTS 2P 2E (Configuration 2)
   case 5002595:                //  DM RTS 2P 1E (Configuration 1) DB9
   case 5002600:                //  DM RTS 2P 1E (Configuration 1) DB9
      ident.numPorts = 2;
       break;

   default:
       diag_printf("Error - unknown model number %d - setting numports to 1 \n", ident.modelId);
       ident.numPorts = 1;
   }

  HAS_ONLY_232 = ModelIsRs232Only();
    
  diag_printf("Heap start:0x%x end:0x%x size:0x%x (%d bytes)\n", 
	      HEAP_START, HEAP_END, HEAP_SIZE, 
	      HEAP_SIZE);

  // lookup port handles and init ownership semaphore
  memset(devMastPort, 0, sizeof devMastPort);
  for (i=0; i<ident.numPorts; ++i)
    {
      Cyg_ErrNo status;
      char portName[32];
      sprintf(portName,"/dev/ser%d",i);
      status = cyg_io_lookup(portName, &devMastPort[i].handle);
      if (status != ENOERR)
        diag_printf("ERROR: cyg_io_lookup %s returned %d\n", portName, status);
      cyg_semaphore_init(&devMastPort[i].ownership,1);
    } 

  /*  
   *  Background task needs to be high priority so we keep kicking the watchdog 
   *  timer even when things are busy...
   */
  cyg_thread_create(1, BgTask, (cyg_addrword_t)0, "Bg",
                    (void *)BgStack, sizeof BgStack, 
                    &BgHandle, &BgThread);
  cyg_thread_resume(BgHandle);
  
  cyg_thread_create(30, DiagTask, (cyg_addrword_t)0, "Diag",
                    (void *)DiagStack, sizeof DiagStack,
                    &DiagHandle, &DiagThread);
  cyg_thread_resume(DiagHandle);

  /*  
   *  Start threads here that want to be started before the scheduler and TCP stack have started.  
   */
  madminInit();
  
  return;  /* returning from this function starts scheduler */
}


/*
 *  Background task.  Toggles LED a bit, starts the TCP Stack 
 *  and watchdog.  Starts all other threads that want to be started
 *  after scheduler and TCP have started.  Then goes into a watchdog strobe loop, 
 */
void BgTask(cyg_addrword_t data)
{
    int i;
    int loopCount = 0;
    
  /* Flash LED a couple times, then leave it on */
  LedOff();
  LedEnable();
  for (i=0; i<15; ++i)
    {
      LedToggle();
      cyg_thread_delay(5);
    }
  LedOn();

  /*  Start stack and watchdog */
  TcpIpStackInit();
  watchdogStart();
  watchdogReset();

  /*  
   *  Start threads here that want to be started after the scheduler and TCP stack have started.  
   */
  tadminInit();
  snmpInit();
  serverInit();
  webServerInit();
  telnetsrvInit();
  portmonInit();

  // enable debug port pins
  PioMode |= Bit(11)+Bit(12)+Bit(13)+Bit(14);

//??
  //  Init the task monitor struct
  taskList[TASK_ETHER_RX].taskNumber = TASK_ETHER_RX;   
  strcpy(taskList[TASK_ETHER_RX].taskName, "EtherRx");  
  taskList[TASK_ETHER_RX].awakeCount = 0;   
  taskList[TASK_ETHER_RX].lastCount = 0; 

  taskList[TASK_TCP_RX].taskNumber = TASK_TCP_RX;       
  strcpy(taskList[TASK_TCP_RX].taskName, "TcpRx");    
  taskList[TASK_TCP_RX].awakeCount = 0;     
  taskList[TASK_TCP_RX].lastCount = 0;

  taskList[TASK_RPSH_POLL].taskNumber = TASK_RPSH_POLL; 
  strcpy(taskList[TASK_RPSH_POLL].taskName, "RpshPoll"); 
  taskList[TASK_RPSH_POLL].awakeCount = 0;  
  taskList[TASK_RPSH_POLL].lastCount = 0;

  taskList[TASK_END_LIST].taskNumber = 0;

  while (1)
    {
      /* Strobe the watchdog every 1/2 second, each delay is 10ms */  	
      watchdogReset();
      cyg_thread_delay(50);  
     
     //  Check for hungs tasks every 5 seconds
      if ((++loopCount % 5) == 0) 
        {
	  i = 0;
	  while (taskList[i].taskNumber) 
            {
	      if (taskList[i].awakeCount == taskList[i].lastCount) 
                diag_printf("Error: Task %s not responding ...\n", taskList[i].taskName);
	      taskList[i].lastCount = taskList[i].awakeCount;
	      i++;
            }
        }
    }
}

void handleDiagCmd(char c)
{
  struct diag_tab_entry *diag_entry;
  
  for (diag_entry = __DeviceMaster_DIAG_TAB__;
       diag_entry != &__DeviceMaster_DIAG_TAB_END__;
       diag_entry++)
    if (diag_entry->cmd == c)
      diag_entry->fun();
}

/*
 *  The diag task is a low priority task that chirps the LED every 10 seconds
 *  as a keepalive status.  Gathers debug header/portmon characters entered.  
 *  Prints an idle count on the debug header/portmon, as a relative indication
 *  of processor busy-ness.
 */
void DiagTask(cyg_addrword_t data)
  {
    extern int portmon_read_char_nonblocking(char *c);
    extern int hal_diag_read_char_nonblocking(char *c);
    int loops;
    extern volatile unsigned idle_thread_loops;
    unsigned long idleLast = idle_thread_loops;
    int i;
    char cmd;
    
    loops = 0;
    
    while (1)
      {
        cyg_thread_delay(99);
        while (hal_diag_read_char_nonblocking(&cmd))
          handleDiagCmd(cmd);
        while (portmon_read_char_nonblocking(&cmd))
          handleDiagCmd(cmd);
        ++loops;
        if (loops % 10 == 0)
          {
            diag_printf("%d: idle=%d  ",loops,(idle_thread_loops-idleLast)/10000);
            idleLast = idle_thread_loops;
            for (i=0; i<ident.numPorts; ++i)
              {
                int semVal;
                cyg_semaphore_peek(&devMastPort[i].ownership,&semVal);
                if (!(i&0x3))
                  diag_printf(" ");
                diag_printf("%d",semVal);
              }
            diag_printf("\n");
            LedOff();
            cyg_thread_delay(10);
            LedOn();
          }
      }
}


#include <network.h>

extern void _show_all_interfaces(void);

//----------------------------------------------------------------------
//
// TCP/IP stack initialization routine.
//
// This routine should be called after the scheduler is started.
// If this fails for any reason, it is retried 3 times.  Then the board is reset.
static void TcpIpStackInit(void)
{
  int s,rc;
  struct sockaddr_in *addrp;
  struct ifreq ifr;
  struct ecos_rtentry route;
  struct in_addr addr;
#ifdef NET_DEBUG
  extern int net_debug;
  net_debug = 1;
#endif

  if (ident.ipAddr == 0xffffffff)   
    return; // not gonna do ip
  
  if (ident.ipAddr == 0x00000000) 
    {
      init_all_network_interfaces(); // use dhcp
      TcpStackOk = 1;
      return;
    }
  
  // otherwise use static address
  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0)
    {
      diag_printf("socket: %s\n", strerror(errno));
      diag_printf("IP network initialization failed\n");
      return;
    }

  strcpy(ifr.ifr_name, "eth0");
  
  addrp = (struct sockaddr_in *) &ifr.ifr_addr;

  memset(addrp, 0, sizeof(*addrp));
  addrp->sin_family = AF_INET;
  addrp->sin_len = sizeof(*addrp);
  addrp->sin_port = 0;

  addrp->sin_addr.s_addr = ident.ipAddr;
  if ((rc=ioctl(s, SIOCSIFADDR, &ifr)))
    diag_printf("Error setting IP address (SIOCSIFADDR): %s\n", strerror(errno));

  addrp->sin_addr.s_addr = ident.netMask;
  if ((rc=ioctl(s, SIOCSIFNETMASK, &ifr)))
    diag_printf("Error setting net mask (SIOCSIFNETMASK): %s\n", strerror(errno));

  addrp->sin_addr.s_addr = ident.ipAddr;
  if ((rc=ioctl(s, SIOCSIFADDR, &ifr)))
    diag_printf("Error setting IP address (SIOCSIFADDR #2): %s\n", strerror(errno));

  addrp->sin_addr.s_addr =  ~ident.netMask | ident.ipAddr;
  if ((rc=ioctl(s, SIOCSIFBRDADDR, &ifr)))
    diag_printf("Error setting broadcast addr (SIOCSIFBRDADDR): %s\n", strerror(errno));

  // Set up routing
  addr.s_addr = ident.ipAddr;
  diag_printf("IP Addr: %s ",inet_ntoa(addr));
  addr.s_addr = ident.netMask;
  diag_printf("Net Mask: %s ",inet_ntoa(addr));
  addr.s_addr = ident.gateway;
  diag_printf("Gateway: %s \n",inet_ntoa(addr));

  memset(&route, 0, sizeof(route));
  
  addrp->sin_addr.s_addr = 0;
  memcpy(&route.rt_dst, addrp, sizeof(*addrp));
  memcpy(&route.rt_genmask, addrp, sizeof(*addrp));
  
  addrp->sin_addr.s_addr = ident.gateway;
  memcpy(&route.rt_gateway, addrp, sizeof(*addrp));
  
  route.rt_dev = ifr.ifr_name;
  route.rt_flags = RTF_UP|RTF_GATEWAY;
  route.rt_metric = 0;
  
  if ((rc=ioctl(s, SIOCADDRT, &route))) 
    diag_printf("Error setting default route (SIOCADDRT): %s\n", strerror(errno));

  TcpStackOk = 1;
  close(s);
}

unsigned long myIpMask(void)
{
  int s;
  unsigned long r;
  struct ifreq ifr;
  
  if ((s=socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    return 0;

  strcpy(ifr.ifr_name, "eth0");

  ifr.ifr_addr.sa_family = AF_INET;
  
  if (ioctl(s, SIOCGIFNETMASK, &ifr))
    r = 0;
  else
    r = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr;

  close(s);
  return r;
}

unsigned long myIpAddr(void)
{
  int s;
  unsigned long r;
  struct ifreq ifr;
  
  if ((s=socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    return 0;

  strcpy(ifr.ifr_name, "eth0");
  
  ifr.ifr_addr.sa_family = AF_INET;
  ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr = -1;
  
  if (ioctl(s, SIOCGIFADDR, &ifr))
      r = 0;
  else
      r = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr;

  close(s);
  return r;
}

#if defined(CYGPKG_NET_FREEBSD_STACK)
#define rtalloc1 cyg_rtalloc1
#endif

#include <sys/param.h>
#include <net/route.h>
extern struct rtentry *rtalloc1(struct sockaddr *dst, int report, int ignflags);

unsigned long myIpGate(void)
{
  struct rtentry *rt;
  struct sockaddr dst;
  memset(&dst,0,sizeof dst);
  dst.sa_family = PF_INET;
  
  rt = rtalloc1(&dst,0, 0);
      
  if (rt && rt->rt_gateway)
    return ((struct sockaddr_in*)(rt->rt_gateway))->sin_addr.s_addr;
  else
    return 0;
}


static void showId(void)
{
  diag_printf("%s\n",versionString);
  diag_printf("Copyright Comtrol Corp. 2001-2004\n");
  diag_printf("Build date: %s\n\n",linkTimeStamp);
}
DeviceMaster_diag(showId,'v',"Show version info");

static void showHelp(void)
{
  struct diag_tab_entry *diag_entry;
  for (diag_entry = __DeviceMaster_DIAG_TAB__;
       diag_entry != &__DeviceMaster_DIAG_TAB_END__;
       diag_entry++)
    diag_printf("%c  %s\n",diag_entry->cmd,diag_entry->str);
}
DeviceMaster_diag(showHelp,'?',"Show available commands");

#if defined(CYGPKG_NET_FREEBSD_STACK)
DeviceMaster_diag(_show_all_interfaces,'I',"Show network interfaces");
#endif

#if defined(CYGPKG_IO_FILEIO)
// don't have cyg_select_abort, so just force a timeout every second

int cyg_select_with_abort(int nsel, fd_set *rfd, fd_set *wfd, fd_set *efd, struct timeval *tv)
{
  struct timeval mytv;
  if (tv == NULL)
    {
      mytv.tv_sec = 1;
      mytv.tv_usec = 0;
      tv = &mytv;
    }
  return select(nsel,rfd,wfd,efd,tv);
}

void cyg_select_abort(void)
{
}
#endif

//  Returns 1 if the model has RS232 only interface, else 0.
//  Serial Hub 8 and 16 port boxes have only RS232.
int ModelIsRs232Only(void)
{
    if (ident.modelId == 5002115 || ident.modelId == 5002315)
	return 1;
    else
	return 0;
}


//  Returns 1 if the model is an Air (Wireless), else 0.
int ModelIsAir(void)
{
    if (ident.modelId == 5002283)
	return 1;
    else
	return 0;
}

//  Returns 1 if the model has RI, else 0.
int ModelHasRI(void)
{
    if ((ident.modelId == 5002280) ||    // DM1
	(ident.modelId == 5002281) ||    // DM1E
	(ident.modelId == 5002111) ||    // DM4 DB9
	(ident.modelId == 5002110) ||    // DM8 DB9
	(ident.modelId == 5002330) ||    // DM8 PRO DB9
	(ident.modelId == 5002283) ||    // DM8 Air 1
	(ident.modelId == 5002115) ||    // DM SH 8
	(ident.modelId == 5002315))      // DM SH 16
	return 1;
    else
	return 0;
}

